from torch.nn import Module, Linear, LogSoftmax,MaxPool2d, ReLU, Conv2d


class LeNet(Module):
	def __init__(self, numChannels, classes):
		# call the parent constructor
		super(LeNet, self).__init__()
		# initialize first set of CONV => RELU => POOL layers
		self.conv1 = Conv2d(in_channels=numChannels, out_channels=20,
			kernel_size=(5, 5))
		self.relu1 = ReLU()
		self.maxpool1 = MaxPool2d(kernel_size=(2, 2), stride=(2, 2))
		# initialize second set of CONV => RELU => POOL layers
		self.conv2 = Conv2d(in_channels=20, out_channels=50,
			kernel_size=(5, 5))
		self.relu2 = ReLU()
		self.maxpool2 = MaxPool2d(kernel_size=(2, 2), stride=(2, 2))
		# initialize first (and only) set of FC => RELU layers
		self.fc1 = Linear(in_features=800, out_features=500)
		self.relu3 = ReLU()
		# initialize our softmax classifier
		#self.fc2 = Linear(in_features=500, out_features=classes)
		#self.logSoftmax = LogSoftmax(dim=1)


from torchsummary import summary
import onnx
from onnx2pytorch import ConvertModel
import torch
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np

__all__ = ['get_model_summary','onnx2torch','overwrite_layer_weights','feature_extractor']


def get_model_summary(model, input_size):
    """Prints the model summary.
    Input:
        model: PyTorch model
        input_size: input size of the model
    """
    summary(model, input_size=input_size)
    
def onnx2torch(model):
    """
    Function to convert onnx model to torch model
    Input:
        model: onnx model
    Output:
        torch model
    """
    onnx_m = onnx.load(model)
    torch_m = ConvertModel(onnx_m)
    return torch_m


def overwrite_layer_weights(model, layer_index, new_weights,logger=None):
    """
    Overwrites the weights of a specified layer of a PyTorch model with the given weights.
    Args:
    - model: A PyTorch model.
    - layer_index: The index of the layer whose weights should be overwritten.
    - new_weights: A tensor containing the new weights to be used for the specified layer.
    """

    layer_name = list(model.named_modules())[layer_index][0]
    layer = getattr(model, layer_name)
    if logger:
        logger.info("Overwriting the weights of layer {} with the given weights.".format(layer_name))
    if isinstance(layer, torch.nn.Conv2d):
        layer.weight.data = new_weights
    if isinstance(layer, torch.nn.Linear):
        layer.weight.data = new_weights
    else:
        raise ValueError("The specified layer is not a convolutional layer or linear layer.")



def feature_extractor(model, layer_name):
    """
    Function to get the feature extractor from the model.
    Args:
        model (torch.nn.Module): PyTorch model
        layer_name (str): Name of the layer to be used as feature extractor
    Returns:
        feature_extractor (torch.nn.Sequential): Feature extractor from the model
    """
    # Get the named module from the model
    module = dict(model.named_modules())[layer_name]
    # Get the position of the named module in the list of all modules
    layer_loc = list(model.modules()).index(module)
    # Extract the feature extractor from the model
    feature_extractor = torch.nn.Sequential(*list(model.children())[:layer_loc])
    return feature_extractor


def feature_view(data,model,layer_names:list) -> None:
    """
    Function to view image from the feature extractor
        data: Image data
        model: PyTorch model
        layer_names (list): Name of the layer to be used as feature extractor
    Output:
        feature_view: Feature view from the model
    """
    out_list = []
    for j in data:
        for i in range(len(layer_names)):
            layer = layer_names[i]
            features = feature_extractor(model,layer)
            #features = getattr(model,layer)
            output = features(j)
            output = output.squeeze(0).detach().numpy()
            out_list.append(output)
        fig, axs = plt.subplots(1, len(layer_names), figsize=(12, 12))
        for i in range(len(layer_names)):
            if len(out_list[i].shape) > 2:
                out_map = np.sum(out_list[i], axis=0)
            else:
                out_map = out_list[i]
            sns.heatmap(out_map, ax=axs[i])
            axs[i].set_title(f'{layer_names[i]} : {out_list[i].shape}' )
    return None
        


import numpy as np
from scipy import spatial

__all__ = ['dice_coefficient_image','numeric_score']

def dice_coefficient_image(pred, target,empty_score=1.0):
    """
    Calculate the dice coefficient between two images
    Formula: 2 * intersection / (pred + target)
    Input:
        pred: numpy array of shape (H, W)
        target: numpy array of shape (H, W)
        empty_score: float
    Output:
        dice_coefficient: float
    """
    
    im1 = np.asarray(pred).astype(np.bool)
    im2 = np.asarray(target).astype(np.bool)
    
    if im1.shape != im2.shape:
        raise ValueError("Shape mismatch: im1 and im2 must have the same shape.")
    
    im_sum = im1.sum() + im2.sum()
    if im_sum == 0:
        return empty_score
    
    intersection = np.logical_and(im1, im2)    
    return 2. * intersection.sum() / im_sum

def numeric_score(pred, target):
    """
    Compute the numeric score between two images like FP, FN, TP, TN
    Formula: FP, FN, TP, TN
    Input:
        pred: numpy array of shape (H, W)
        target: numpy array of shape (H, W)
    Output:
        numeric_score: list of int
    """
    
    FP = np.logical_and(pred, np.logical_not(target)).sum()
    


#dataloader for pytorch1.0

import torch
import torchvision
import torchvision.transforms as transforms
from torch.utils.data import DataLoader
from mb_pytorch.utils.yaml_reader import YamlReader
import os
import sys
import numpy as np
from mb_pandas.src.dfload import load_any_df
from mb_utils.src.verify_image import verify_image
from mb_pandas.src.transform import *
from datetime import datetime
from PIL import Image
import cv2

today = datetime.now()

__all__ = ['data_fetcher','DataLoader']

class data_fetcher:
    """
    dataloader for pytorch1.0
    """
    def __init__(self,yaml,logger=None) -> None:
        self.yaml = yaml
        self.logger = logger
        self._yaml_data = None
        self.data_dict = {}
        self.transforms_final=[]
        self.all = None

    def __repr__(self) -> str:
        return "data_fetcher(yaml={},logger={})".format(self.yaml,self.logger)

    @staticmethod
    def read_yaml(self):
        """
        read yaml file
        """
        self._yaml_data = YamlReader(self.yaml).data(self.logger)
        return self._yaml_data

    @property
    def load_data_params(self):
        """
        get dataloader data from yaml file
        """
        data = YamlReader(self.yaml).data(self.logger)
        self.data_dict['data'] = data['data']
        self.data_dict['train_params'] = data['train_params']
        self.data_dict['test_params'] = data['test_params']
        self.data_dict['transforms_list'] = data['transformation']
        self.data_dict['model'] = data['model']
        return self.data_dict
    
    @property
    def load_data_all(self):
        """
        get dataloader all data dict from yaml file
        """
        data = YamlReader(self.yaml).data(self.logger)
        self.all = data
        return self.all

    @property
    def get_transforms(self):
        """
        get transforms from yaml file
        """

        transforms_list = self.load_data_params['transforms_list']

        if transforms_list['transform']==False:
            return None

        # for t_list in transforms_list:
        #     if t_list in dir(transforms):
        #         self.transforms_final.append(transforms.t_list)

        if 'transforms' in dir(self.transforms_final):
            self.transforms_final = []

        if len(self.transforms_final) != 0:
            self.transforms_final = []

        #for t_list in transforms_list:
        if transforms_list['to_tensor']['val']:
            self.transforms_final.append(transforms.ToTensor())
        if transforms_list['normalize']['val']:
            self.transforms_final.append(transforms.Normalize(transforms_list['normalize']['args']['mean'],transforms_list['normalize']['args']['std']))
        if transforms_list['resize']['val']:
            self.transforms_final.append(transforms.Resize(transforms_list['resize']['args']['size']))
        if transforms_list['random_crop']['val']:
            self.transforms_final.append(transforms.RandomCrop(transforms_list['random_crop']['args']['size']))
        if transforms_list['random_horizontal_flip']['val']:
            self.transforms_final.append(transforms.RandomHorizontalFlip(transforms_list['random_horizontal_flip']['args']['p']))
        if transforms_list['random_vertical_flip']['val']:
            self.transforms_final.append(transforms.RandomVerticalFlip(transforms_list['random_vertical_flip']['args']['p']))
        if transforms_list['random_rotation']['val']:
            self.transforms_final.append(transforms.RandomRotation(transforms_list['random_rotation']['args']['degrees']))
        if transforms_list['random_color_jitter']['val']:
            self.transforms_final.append(transforms.ColorJitter(transforms_list['random_color_jitter']['args']['brightness'],transforms_list['random_color_jitter']['args']['contrast'],transforms_list['random_color_jitter']['args']['saturation'],transforms_list['random_color_jitter']['args']['hue']))
        if transforms_list['random_grayscale']['val']:
            self.transforms_final.append(transforms.RandomGrayscale(transforms_list['random_grayscale']['args']['p']))
        if self.logger:
            self.logger.info("transforms: {}".format(self.transforms_final))
        
        self.transforms_final = transforms.Compose(self.transforms_final)
        return self.transforms_final
   
class customdl(torch.utils.data.Dataset):
    def __init__(self,data,transform=None,train_file=True,logger=None):
        self.transform=transform
        self.logger=logger
        self.folder_name=data['work_dir']
        self.data = load_any_df(data['file'],logger=self.logger)

        if self.logger:
            self.logger.info("Data file: {} loaded with mb_pandas.".format(data))
            self.logger.info("Data columns: {}".format(self.data.columns))
            self.logger.info("Data will be split into train and validation according to train_file input : {}".format(train_file))
            self.logger.info("If unnamed columns are present, they will be removed.")
            self.logger.info("If duplicate rows are present, they will be removed.")
        assert 'image_path' in self.data.columns, "image_path column not found in data"
        assert 'image_type' in self.data.columns, "image_type column not found in data"

        if train_file:
            self.data = self.data[self.data['image_type'] == 'training']
        else:
            self.data = self.data[self.data['image_type'] == 'validation']

        if 'label' in self.data.columns:
            self.label = self.data['label']
        else:
            self.label = None

        self.data = check_drop_duplicates(self.data,columns=['image_path'],drop=True,logger=self.logger)
        self.data = remove_unnamed(self.data,logger=self.logger)

        # else:
        #     date_now = today.strftime("%d_%m_%Y_%H_%M")
        #     self.folder_name='data_'+date_now
        # os.mkdir('./data'+str(self.folder_name))

        if data['use_img_dir']:
            img_path = [os.path.join(str(data['img_dir']),self.data['image_path'].iloc[i]) for i in range(len(self.data))]
        else:
            img_path = [self.data['image_path'].iloc[i] for i in range(len(self.data))]
        self.data['image_path_new'] = img_path
        if self.logger:
            self.logger.info("Verifying paths")
            self.logger.info("first path : {}".format(img_path[0]))

        path_check_res= [os.path.exists(img_path[i]) for i in range(len(img_path))]
        self.data['img_path_check'] = path_check_res
        self.data = self.data[self.data['img_path_check'] == True]
        self.data = self.data.reset_index(drop=True)
        if logger:
            self.logger.info("self.data: {}".format(self.data))

        if data['thresholding_pd']>0:
            if len(self.data) <= data['thresholding_pd']:
                self.logger.info("Length of data after removing invalid paths: {}".format(len(self.data)))
                self.logger.info("Less than thresholding_pd data points. Please check the data file.")
                self.logger.info("Exiting")
                sys.exit('Less than thresholding_pd data points. Please check the data file.')

        if self.logger:
            self.logger.info("Length of data after removing invalid paths: {}".format(len(self.data)))
            self.logger.info("Verifying images")
        verify_image_res = [verify_image(self.data['image_path_new'].iloc[i],logger=self.logger) for i in range(len(self.data))]  
        self.data['img_verify'] = verify_image_res
        self.data = self.data[self.data['img_verify'] == True]
        self.data = self.data.reset_index()

        if data['thresholding_pd']>0:
            if len(self.data) <= data['thresholding_pd']:
                self.logger.info("Length of data after removing invalid images: {}".format(len(self.data)))
                self.logger.info("Less than thresholding_pd data points. Please check the data file.")
                self.logger.info("Exiting")
                sys.exit('Less than thresholding_pd data points. Please check the data file.')
        
        if os.path.exists(self.folder_name):
            self.data.to_csv(os.path.join(self.folder_name,'wrangled_file.csv'),index=False)

    def __len__(self):
        return len(self.data)
    
    def __repr__(self) -> str:
        return "self.data: {},self.transform: {},self.label: {}".format(self.data,self.transform,self.label)

    def __getitem__(self,idx):
        
        img = self.data['image_path_new'].iloc[idx]
        #img = Image.open(img)
        img = cv2.imread(img)
        if self.label:
            label = self.label[idx]
        
        if self.transform:
            img = self.transform(img)

        out_dict = {'image':img}
        if self.label:
            out_dict['label'] = label                                                
        return out_dict

class DataLoader(data_fetcher):
    """
    Basic dataloader for pytorch1.0
    """
    def __init__(self,yaml,logger=None) -> None:
        super().__init__(yaml, logger=logger)
        self.yaml = yaml
        self.logger = logger
        self._yaml_data = None
        self.data_dict = self.load_data_params
        self.transforms_final=[]
        self.trainloader = None
        self.testloader = None
        self.folder_name = self.data_dict['data']['work_dir']
        self.data_file= self.data_dict['data']['from_datasets']

        if os.path.exists(self.folder_name):
            if self.logger:
                self.logger.info("Data folder already exists. Using existing data folder :  {}".format(self.folder_name))
        else:
            os.mkdir(self.folder_name)
            if self.logger:
                self.logger.info("Data folder created : {}".format(self.folder_name))
    
    def data_load(self):
        """
        return all data loaders
        """

        if self.data_dict['data']['from_file']==False:
            if self.data_file in dir(torchvision.datasets):
                if self.logger:
                    self.logger.info("Data file: {} loading from torchvision.datasets.".format(self.data_file))
                if self.data_file in os.listdir(self.folder_name):
                    download_flag = False
                else:
                    download_flag = True
                if self.data_file == 'CIFAR10' or self.data_file == 'CIFAR100':
                    self.trainset = getattr(torchvision.datasets,self.data_file)(root=self.folder_name, train=True, download=download_flag,transform=self.get_transforms)
                    self.testset = getattr(torchvision.datasets,self.data_file)(root=self.folder_name, train=False, download=download_flag,transform=self.get_transforms)
                else:
                    self.trainset = getattr(torchvision.datasets,self.data_file)(root=self.folder_name, split='train', download=download_flag,transform=self.get_transforms)
                    self.testset = getattr(torchvision.datasets,self.data_file)(root=self.folder_name, split='val', download=download_flag,transform=self.get_transforms)
                if self.data_dict['data']['thresholding_pd']>0:
                    subset_indices = range(self.data_dict['data']['thresholding_pd'])
                    self.trainset = torch.utils.data.Subset(self.trainset, subset_indices)
                    self.testset = torch.utils.data.Subset(self.testset, subset_indices)
            else:
                if self.logger:
                    self.logger.info("Data file: {} could not be loaded from torchvision.datasets.".format(self.data_file))
                    self.logger.info("Exiting")
                sys.exit("Data file: {} could not be loaded from torchvision.datasets.".format(self.data_file))
        else:
            self.trainset = self.data_train(self.data_dict['data'],transform=self.get_transforms,train_file=True,logger=self.logger)
            self.testset = self.data_train(self.data_dict['data'],transform=self.get_transforms,train_file=False,logger=self.logger)

        self.trainloader = torch.utils.data.DataLoader(self.trainset, batch_size=self.data_dict['train_params']['batch_size'], shuffle=self.data_dict['train_params']['shuffle'], num_workers=self.data_dict['train_params']['num_workers'],worker_init_fn = lambda id: np.array(self.data_dict['train_params']['seed']))
        self.testloader = torch.utils.data.DataLoader(self.testset, batch_size=self.data_dict['test_params']['batch_size'], shuffle=self.data_dict['test_params']['shuffle'], num_workers=self.data_dict['test_params']['num_workers'],worker_init_fn = lambda id: np.array(self.data_dict['test_params']['seed']))
        return self.trainloader,self.testloader,self.trainset,self.testset

    def data_train(self,data_file,transform,train_file,logger=None):
        """
        get train data from yaml file
        """
        data_t = customdl(data_file,transform=transform,train_file=train_file,logger=logger)
        return data_t


import matplotlib.pyplot as plt
import numpy as np
import cv2
import torch
import torchvision
from torchvision.utils import draw_bounding_boxes, draw_segmentation_masks
import torchvision.transforms.functional as TF
from pytorch_grad_cam import GradCAM
from pytorch_grad_cam.utils.image import show_cam_on_image
import io
import PIL

__all__ = ['show_images', 'show_segmentation_masks', 'show_bounding_boxes', 'show_label_on_img','model_viewer','new_show_cam_on_image','gradcam_viewer','plot_classes_pred']

def show_images(imgs, figsize=(12.0, 12.0)):
    """Displays a single image or list of images. 
    Args:
        imgs (Union[List[torch.Tensor], torch.Tensor]): A list of images
            of shape (3, H, W) or a single image of shape (3, H, W).
        figsize (Tuple[float, float]): size of figure to display.
    Returns:
        None
    """

    if not isinstance(imgs, list):
        imgs = [imgs]
    fig, axs = plt.subplots(ncols=len(imgs), figsize=figsize, squeeze=False)
    for i, img in enumerate(imgs):
        #img = img.detach()
        #img = TF.to_pil_image(img)
        img = np.asarray(img)
        if img.shape[0]==3:
            img = np.transpose(img, (1, 2, 0))
        axs[0, i].imshow(np.asarray(img))
        axs[0, i].set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])
    plt.show()

    return None
    
def plot_to_image(figure):
    """Converts the matplotlib plot specified by 'figure' to a PNG image and
    returns it. The supplied figure is closed and inaccessible after this call.
    Source: (https://www.tensorflow.org/tensorboard/image_summaries)."""
    # Save the plot to a PNG in memory.
    buf = io.BytesIO()
    figure.savefig(buf, format='png')
    buf.seek(0)
    image = PIL.Image.open(buf)  
    img_tensor = torch.tensor(np.array(image)).permute(2, 0, 1).float() / 255.0      
    #img_ar = np.array(image)
    #image = torch.Tensor(img_ar)
    return img_tensor

def create_img_grid(data,labels, t_writer, n_images=9, g_i=3, g_j=3,global_step=0,gray_image=False):
    """
    Creates a grid of images and sends it to TensorBoard.
    Input:
        data: A list of images
        labels: A list of labels
        t_writer: TensorBoard writer
        n_images: Number of images to display
        g_i: Number of rows in the grid
        g_j: Number of columns in the grid
    """
    
    data_read = data[:n_images]
    labels_read = labels[:n_images]
    
    ### Visualize a 3x3 grid
    fig, axes = plt.subplots(g_i, g_j, figsize=(15,15),
            subplot_kw = {'xticks':[], 'yticks':[]},
            gridspec_kw = dict(hspace=0.3, wspace=0.01))
    
    if gray_image:
        for i, ax in enumerate(axes.flat):
            img_cur = np.array(data_read[i,0,:,:])
            ax.title.set_text(int(labels_read[i]))
            ax.imshow(img_cur, cmap='gray')
    else:
        for i, ax in enumerate(axes.flat):
            img_cur = np.array(data_read[i])
            img_cur = np.transpose(img_cur, (1, 2, 0))
            ax.title.set_text(int(labels_read[i]))
            ax.imshow(img_cur)
    
    ### Send the figure over to TensorBoard
    t_writer.add_image('grid', plot_to_image(fig), global_step=global_step)
    plt.close(fig)

    

def show_segmentation_masks(imgs, masks, figsize=(12.0, 12.0)):
    """Displays a single image or list of images with segmentation masks.
    Args:
        imgs (Union[List[torch.Tensor], torch.Tensor]): A list of images
            of shape (3, H, W) or a single image of shape (3, H, W).
        masks (Union[List[torch.Tensor], torch.Tensor]): A list of masks
            of shape (1, H, W) or a single mask of shape (1, H, W).
        figsize (Tuple[float, float]): size of figure to display.
    Returns:
        None
    """
    if not isinstance(imgs, list):
        imgs = [imgs]
    if not isinstance(masks, list):
        masks = [masks]
    fig, axs = plt.subplots(ncols=len(imgs), figsize=figsize, squeeze=False)
    for i, (img, mask) in enumerate(zip(imgs, masks)):
        # img = img.detach()
        # img = TF.to_pil_image(img)
        # mask = mask.detach()
        # mask = TF.to_pil_image(mask)
        img = np.asarray(img)
        if img.shape[0]==3:
            img = np.transpose(img, (1, 2, 0))
        mask = np.asarray(mask)
        if mask.shape[0]==3:
            mask = np.transpose(mask, (1, 2, 0))
        axs[0, i].imshow(np.asarray(img))
        axs[0, i].imshow(np.asarray(mask), alpha=0.5)
        axs[0, i].set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])
    plt.show()

    return None

def show_bounding_boxes(imgs, boxes, figsize=(12.0, 12.0)):
    """Displays a single image or list of images with bounding boxes.
    Args:
        imgs (Union[List[torch.Tensor], torch.Tensor]): A list of images
            of shape (3, H, W) or a single image of shape (3, H, W).
        boxes (Union[List[torch.Tensor], torch.Tensor]): A list of boxes
            of shape (N, 4) or a single box of shape (N, 4).
        figsize (Tuple[float, float]): size of figure to display.
    Returns:
        None
    """
    if not isinstance(imgs, list):
        imgs = [imgs]
    if not isinstance(boxes, list):
        boxes = [boxes]
    fig, axs = plt.subplots(ncols=len(imgs), figsize=figsize, squeeze=False)
    for i, (img, box) in enumerate(zip(imgs, boxes)):
        # img = img.detach()
        # img = TF.to_pil_image(img)
        # box = box.detach()
        img = np.asarray(img)
        if img.shape[0]==3:
            img = np.transpose(img, (1, 2, 0))
        box = np.asarray(box)
        axs[0, i].imshow(np.asarray(img))
        axs[0, i].imshow(draw_bounding_boxes(img, box))
        axs[0, i].set(xticklabels=[], yticklabels=[], xticks=[], yticks=[])
    plt.show()

    return None

def show_label_on_img(imgs,labels,figsize=(12.0,12.0)):
    """Displays a single image or list of images with labels.
    Args:
        imgs (Union[List[torch.Tensor], torch.Tensor]): A list of images
            of shape (3, H, W) or a single image of shape (3, H, W).
        labels (str): label to be displayed on image.
        figsize (Tuple[float, float]): size of figure to display.
    Returns:
        None
    """
    if not isinstance(imgs, list):
        imgs = [imgs]
    if not isinstance(labels, list):
        labels = [labels]
    fig, axs = plt.subplots(ncols=len(imgs), figsize=figsize, squeeze=False)
    for i, (img, label) in enumerate(zip(imgs, labels)):
        # img = img.detach()
        # img = TF.to_pil_image(img)
        img = np.asarray(img)
        if img.shape[0]==3:
            img = np.transpose(img, (1, 2, 0))
        img = img.astype('float32')
        img = cv2.cvtColor(img, cv2.COLOR_BGR2RGB)
        cv2.putText(img, str(label), (10, 30), cv2.FONT_HERSHEY_SIMPLEX, 0.25, (0, 0, 255), 1)
        axs[0, i].imshow(np.asarray(img))
    plt.show()

    return None

def model_viewer(model,input_shape = (1,3,128,128), location='model',view=False):
    """
    Function to visualize the model architecture
    Uses torchviz to visualize the model
    Input:
        model: Model to be visualized
        input_shape: Input shape of the model (batch,channels,height,width)
        location: name and location to save the model image. Default is current directory
        view: Boolean to view the model image or not
    Output:
        None
    """
    from torchviz import make_dot
    x = torch.randn(input_shape)
    y = model(x)
    dot= make_dot(y.mean(), params=dict(model.named_parameters()),show_attrs=True)
    dot.format = 'png'
    dot.render(location,view=view)
    return None
    
def new_show_cam_on_image(img, mask, use_rgb=True,image_weight: float = 0.4):
    """
    Input:
        img: Image on which the cam is to be superimposed
        mask: cam mask
    Return:
        cam: superimposed image
    """
    #mask = cv2.cvtColor(mask, cv2.COLOR_BGR2GRAY) # Convert to single channel
    mask = np.float32(mask) / 255 # Convert to float32
    heatmap = cv2.applyColorMap(np.uint8(255 * mask), cv2.COLORMAP_JET) # Convert to uint8
    if use_rgb:
        heatmap = cv2.cvtColor(heatmap, cv2.COLOR_BGR2RGB)
    heatmap = np.float32(heatmap) / 255

    # if np.max(img) > 1:
    #     raise Exception(
    #         "The input image should np.float32 in the range [0, 1]")
    if image_weight < 0 or image_weight > 1:
        raise Exception(
            f"image_weight should be in the range [0, 1].\
                Got: {image_weight}")
        
    #if use_rgb:
    #    cam = heatmap[..., ::-1] + np.float32(img)
    #else:
    #    cam = heatmap[..., 0] * np.float32(img)
    
    if use_rgb:
        cam = (1 - image_weight) * heatmap[..., ::-1] + image_weight * img
    else:
        cam = (1 - image_weight) * heatmap[..., 0] + image_weight * img
    cam = cam / np.max(cam)
    
    return np.uint8(255 * cam)

def gradcam_viewer(gradcam_layer, model, x_grad,gradcam_rgb=False,use_cuda=False):
    split_val = gradcam_layer.split('.')[1]
    new_layer_name = 'model.' + split_val
    gradcam_eval = eval(new_layer_name)
    with GradCAM(model=model,target_layers=[gradcam_eval],use_cuda=use_cuda) as cm: 
        try:
            if split_val == 'classifier' or 'fc':
                cr = cm(input_tensor=x_grad)[0,:]
        except:
            cr = None
            return cr
        if cr is not None:        
            #cam_img = new_show_cam_on_image(x_grad[0].numpy(),cr,use_rgb=gradcam_rgb)
            x_grad_new = x_grad[0].permute(1,2,0).numpy()
            #cam_img = show_cam_on_image(x_grad_new, cr,use_rgb=gradcam_rgb)
            cam_img = new_show_cam_on_image(x_grad_new, cr,use_rgb=gradcam_rgb)
    return cam_img
    

def plot_classes_pred(images, labels, predictions_prob, preds):
    fig = plt.figure(figsize=(20.0, 60.0))
    if type(preds) == torch.Tensor:
        preds = preds.to('cpu')
    if type(labels) == torch.Tensor:
        labels = labels.to('cpu')
    preds_np = preds.numpy()
    predictions_prob_np = predictions_prob.numpy()
    labels_np = labels.numpy()

    for idx in np.arange(4):
        ax = fig.add_subplot(1, 4, idx+1, xticks=[], yticks=[])
        images_permuted = images.permute(0, 2, 3, 1)
        plt.imshow(images_permuted[idx,:])
        ax.set_title("{}, {:.1f}%\n(label: {})".format(
                    str(preds_np[idx]),
                    float(predictions_prob_np[idx].max()) * 100.0,
                    str(labels_np[idx])),
                    color=("green" if preds[idx]==labels[idx] else "red"))
    return fig


import torch
from torch import nn
from .conv_with_relu import ConvBlock


__all__ = ['CNN','CustomCNN']

class CNN(nn.Module):
   """
   Applies several ConvBlocks each doubling the number of channels, and halving the feature map size, before taking a global average and classifying.
   """

   def __init__(self, in_channels, num_blocks, num_classes):
       super().__init__()
       first_channels = 64
       self.blocks = nn.ModuleList(
           [ConvBlock(
               num_layers=3,
               in_channels=(in_channels if i == 0 else first_channels*(2**(i-1))),
               out_channels=first_channels*(2**i))
            for i in range(num_blocks)]
       )
       self.global_pool = nn.AdaptiveAvgPool2d((1, 1))
       self.cls = nn.Linear(first_channels*(2**(num_blocks-1)), num_classes)

   def forward(self, x):
       for block in self.blocks:
           x = block(x)
       x = self.global_pool(x)
       x = x.flatten(1)
       x = self.cls(x)
       return x

class CustomCNN(nn.Module):
    """
    Custom CNN model.
    Inputs:
        depth (int): Number of convolutional layers
        layer_of_extraction (int): Layer of feature extraction
        skip_connection (bool): Whether to use skip connections
    
    """
    def __init__(self, depth, layer_of_extraction, skip_connection):
        super(CustomCNN, self).__init__()
        
        self.depth = depth
        self.layer_of_extraction = layer_of_extraction
        self.skip_connection = skip_connection
        
        self.conv_layers = nn.ModuleList()
        self.pool_layers = nn.ModuleList()
        self.skip_connections = nn.ModuleList()
        
        # Define convolutional layers
        for i in range(depth):
            in_channels = 3 if i == 0 else 64
            out_channels = 64
            kernel_size = 3
            padding = 1
            conv_layer = nn.Conv2d(in_channels, out_channels, kernel_size, padding,stride=1)
            self.conv_layers.append(conv_layer)
            
            # Define skip connections
            if skip_connection and i > 0:
                skip_connection_layer = nn.Conv2d(in_channels, out_channels, 1)
                identity = nn.Identity()
                self.skip_connections_layer.append([skip_connection_layer, identity])
                
            # Define pooling layers
            if i < depth - 1:
                pool_layer = nn.MaxPool2d(2, 2)
                self.pool_layers.append(pool_layer)
        
        # Define fully connected layer
        self.fc_layer = nn.Linear(64 * 8 * 8, 10)
        
    def forward(self, x):
        layer_outputs = []
        
        for i in range(self.depth):
            x = self.conv_layers[i](x)
            
            if self.skip_connection and i > 0:
                skip_connection_layer, skip_connection_identity = self.skip_connections[i - 1]
                if skip_connection_layer is not None:
                    skip_connection_out = skip_connection_layer(x)
                x = x + skip_connection_out
            
            x = nn.functional.relu(x)
            
            if i < self.depth - 1:
                x = self.pool_layers[i](x)
            
            if i == self.layer_of_extraction:
                layer_outputs.append(x)
        
        x = torch.flatten(x, start_dim=1)
        x = self.fc_layer(x)
        
        if self.layer_of_extraction == self.depth-1 and len(layer_outputs) == 0:            
            layer_outputs.append(x)
        elif self.layer_of_extraction < self.depth-1:
            layer_outputs.append(x)
        
        return x, layer_outputs[-1]
